<?php

namespace WpTypek\Platform;

defined( '_WP_EXEC' ) or die( 'Restricted access' );

use WpTypek\Core\Exception;
use WpTypek\Core\Helper;
use WpTypek\Core\PreloadCache;

class Cache implements \WpTypek\Interfaces\CacheInterface {
    protected static $wp_filesystem;

    /**
     *
     * @param string $id
     * @param bool $check_expire
     * @param bool $page_cache
     *
     * @return string|false
     * @throws Exception
     */
    public static function getCache( $id, $check_expire = false, $page_cache = false ) {
        $wp_filesystem = self::getWpFileSystem();

        if ( $wp_filesystem === false ) {
            return false;
        }

        $file = self::_getFileName( $id, $page_cache );

        if ( ! $wp_filesystem->exists( $file ) ) {
            return false;
        }

        if ( $check_expire && time() > $wp_filesystem->mtime( $file ) + self::getLifetime( $page_cache ) ) {
            Utility::cacheLog( $file, 'expired' );

            return false;
        }

        return self::_getCacheFile( $file, $wp_filesystem );
    }

    /**
     *
     * @param string $id
     * @param callable $function
     * @param array $args
     *
     * @return string
     * @throws Exception
     */
    public static function getCallbackCache( $id, $function, $args ) {
        $wp_filesystem = self::getWpFileSystem();

        if ( $wp_filesystem === false ) {
            return false;
        }

        $file = self::_getFileName( $id );

        if ( ! self::getCache( $id, true ) ) {
            $contents = call_user_func_array( $function, $args );

            return self::saveCache( $contents, $id );
        }

        return self::_getCacheFile( $file, $wp_filesystem );
    }

    /**
     *
     * @param string $context
     *
     * @return bool
     * @throws Exception
     */
    public static function deleteCache( $context = 'both' ) {
        Utility::cacheLog( 'deleteCache(' . $context . ')', 'delete' );

        //Purge LiteSpeed cache
        if ( $context != 'plugin' && class_exists( 'LiteSpeed_Cache_API' ) ) {
            \LiteSpeed_Cache_API::purge_all();
        }

        $wp_filesystem = self::getWpFileSystem();

        if ( $wp_filesystem === false || ! $wp_filesystem->exists( WPTYPEK_CACHE_DIR ) ) {
            return false;
        }

        if ( $context == 'plugin' ) {
            $result = true;
            $result |= (bool) $wp_filesystem->rmdir( WPTYPEK_CACHE_DIR . 'js', true );
            $result |= (bool) $wp_filesystem->rmdir( WPTYPEK_CACHE_DIR . 'css', true );

            return $result;
        }

        $cache_dir = dirname( WPTYPEK_CACHE_DIR );
        //Get list of all folders in the cache directory (.../wp-content/cache/)
        $cache_dir_list = $wp_filesystem->dirlist( $cache_dir, false, false );

        foreach ( $cache_dir_list as $entry ) {
            //Skip the wp-typek cache if we're only deleting page cache
            if ( $context == 'page' && $entry['name'] == 'wp-typek' ) {
                $wp_filesystem->rmdir( $cache_dir . '/wp-typek/page', true );
                PreloadCache::clearAllCache();

                continue;
            }


            if ( $entry['type'] == 'd' ) {
                //Delete each cache folder, including WPTypek cache
                if ( ! $wp_filesystem->rmdir( $cache_dir . DIRECTORY_SEPARATOR . $entry['name'], true ) ) {
                    return false;
                }
            }
        }

        // ## <plugin_mod@mirek> ##
        if ( $context === 'both' ) {
            PreloadCache::clearAllCache();
        }

        return true;
    }

    /** ## <plugin_mod@mirek> ## */
    public static function deletePagecache( $pageCacheId ) {
        Cache::getWpFileSystem(); //initiate cache dir
        $file = self::_getFileName( $pageCacheId, true );
        if ( file_exists( $file ) ) {
            @unlink( $file );
        }
        $headerFile = str_replace( '.html', '.header', $file );
        if ( file_exists( $headerFile ) ) {
            @unlink( $headerFile );
        }

        return true;
    }

    /** ## <plugin_mod@mirek> ## */
    public static function deleteCssCacheById( $cssCacheId ) {
        if ( ! empty( $cssCacheId ) ) {
            Cache::getWpFileSystem();
            $cssFile = WPTYPEK_CACHE_DIR . 'css/' . $cssCacheId . '.css';
            if ( file_exists( $cssFile ) ) {
                @unlink( $cssFile );
            }
        }

        return true;
    }

    /** ## <plugin_mod@mirek> ## */
    public static function deleteJsCacheById( $jsCacheId ) {
        if ( ! empty( $jsCacheId ) ) {
            Cache::getWpFileSystem();
            $jsFile = WPTYPEK_CACHE_DIR . 'js/' . $jsCacheId . '.js';
            if ( file_exists( $jsFile ) ) {
                @unlink( $jsFile );
            }
        }

        return true;
    }

    /**
     *
     * @param string $file
     * @param \WP_Filesystem_Base $wp_filesystem
     *
     * @return string
     */
    private static function _getCacheFile( $file, $wp_filesystem ) {
        $content = $wp_filesystem->get_contents( $file );

        //##<plugin_mod@mirek>##
        $unserialized = @unserialize( $content );
        if ( $unserialized !== false ) {
            return $unserialized;
        } else {
            return $content;
        }
    }

    /**
     *
     * @throws Exception
     */
    public static function initializeCache() {
        $wp_filesystem = self::getWpFileSystem();

        if ( $wp_filesystem !== false ) {
            $index_contents = '<html lang=""><body></body></html>';

            if ( ! $wp_filesystem->exists( WPTYPEK_CACHE_DIR ) ) {
                if ( ! $wp_filesystem->exists( $wp_filesystem->wp_content_dir() . 'cache' ) ) {
                    $wp_filesystem->mkdir( $wp_filesystem->wp_content_dir() . 'cache', FS_CHMOD_DIR );
                }

                $wp_filesystem->mkdir( WPTYPEK_CACHE_DIR, FS_CHMOD_DIR );
                $wp_filesystem->put_contents( WPTYPEK_CACHE_DIR . 'index.html', $index_contents, FS_CHMOD_FILE );
            }

            if ( ! $wp_filesystem->exists( WPTYPEK_CACHE_DIR . 'page' ) ) {
                $wp_filesystem->mkdir( WPTYPEK_CACHE_DIR . 'page', FS_CHMOD_DIR );
                $wp_filesystem->put_contents( WPTYPEK_CACHE_DIR . 'page/index.html', $index_contents, FS_CHMOD_FILE );
            }

            /** ## <plugin_mod@mirek> ## */
            if ( ! $wp_filesystem->exists( WPTYPEK_CACHE_DIR . 'page' . DIRECTORY_SEPARATOR . 'mobile' ) ) {
                $wp_filesystem->mkdir( WPTYPEK_CACHE_DIR . 'page' . DIRECTORY_SEPARATOR . 'mobile', FS_CHMOD_DIR );
                $wp_filesystem->put_contents( WPTYPEK_CACHE_DIR . 'page/mobile/index.html', $index_contents, FS_CHMOD_FILE );
            }

            if ( ! $wp_filesystem->exists( WPTYPEK_CACHE_DIR . 'css' ) ) {
                $wp_filesystem->mkdir( WPTYPEK_CACHE_DIR . 'css', FS_CHMOD_DIR );
            }

            if ( ! $wp_filesystem->exists( WPTYPEK_CACHE_DIR . 'js' ) ) {
                $wp_filesystem->mkdir( WPTYPEK_CACHE_DIR . 'js', FS_CHMOD_DIR );
            }
        }
    }

    /**
     *
     * @return \WP_Filesystem_Base|false
     * @throws Exception
     */
    public static function getWpFileSystem() {
        if ( ! isset( self::$wp_filesystem ) ) {
            /** @var \WP_Filesystem_Base $wp_filesystem */
            global $wp_filesystem;

            $wp_filesystem_cache = $wp_filesystem;

            if ( ! function_exists( 'request_filesystem_credentials' ) ) {
                include_once Paths::rootPath() . '/wp-admin/includes/file.php';
            }

            add_filter( 'request_filesystem_credentials', array( 'WpTypek\Platform\Cache', 'requestFilesystemCredentials' ), 10, 7 );

            if ( false === ( $creds = request_filesystem_credentials( admin_url( 'admin.php?page=wp-typek_settings' ), '', false,
                    WP_CONTENT_DIR, null, true ) ) ) {
                $message = 'The plugin needs to access the filesystem via FTP. Please define your FTP credentials in your wp_config.php file like this: <br /><br /><textarea disabled="disabled" rows="3" cols="40" style="background-color: ghostwhite; color: #444; resize: none;">define(\'FTP_HOST\', \'[Your FTP host]\'); &#13;&#10;define(\'FTP_USER\', \'[Your FTP Username]\'); &#13;&#10;define(\'FTP_PASS\', \'[Your FTP password]\'); </textarea>';

                self::$wp_filesystem = false;

                if ( function_exists( 'wptypek_flash_message' ) ) {
                    wptypek_flash_message( 'error', __( $message, 'wp-typek' ) );

                    return false;
                } else {
                    throw new Exception( __( $message, 'wp-typek' ) );
                }
            }

            if ( false === \WP_Filesystem( $creds, WP_CONTENT_DIR, true ) ) {
                $message = 'Could not connect to the filesystem. Please check your FTP credentials in wp_config.php file';

                if ( function_exists( 'wptypek_flash_message' ) ) {
                    wptypek_flash_message( 'error', __( $message, 'wp-typek' ) );

                    return false;
                } else {
                    throw new Exception( __( $message ) );
                }
            }

            self::$wp_filesystem = $wp_filesystem;

            if ( ! defined( 'WPTYPEK_CACHE_DIR' ) ) {
                define( 'WPTYPEK_CACHE_DIR', $wp_filesystem->wp_content_dir() . 'cache/wp-typek/' );
            }

            $wp_filesystem = $wp_filesystem_cache;
        }

        return self::$wp_filesystem;
    }

    /**
     * @param $value
     * @param $form_post
     * @param $type
     * @param $error
     * @param $context
     * @param $extra_fields
     * @param $allow_relaxed_file_ownership
     *
     * @return bool|mixed|void
     */
    public static function requestFilesystemCredentials( $value, $form_post, $type, $error, $context, $extra_fields, $allow_relaxed_file_ownership ) {
        $method = get_filesystem_method( array(), $context, $allow_relaxed_file_ownership );

        if ( $method == 'direct' ) {
            return true;
        }

        if ( ! function_exists( 'wp_generate_password' ) ) {
            include_once Paths::rootPath() . '/wp-includes/pluggable.php';
        }

        $credentials = get_option( 'ftp_credentials', array( 'hostname' => '', 'username' => '' ) );

        $credentials['hostname'] = defined( 'FTP_HOST' ) ? FTP_HOST : $credentials['hostname'];
        $credentials['username'] = defined( 'FTP_USER' ) ? FTP_USER : $credentials['username'];
        $credentials['password'] = defined( 'FTP_PASS' ) ? FTP_PASS : '';

        if ( $method == 'ssh2' ) {
            // Check to see if we are setting the public/private keys for ssh
            $credentials['public_key']  = defined( 'FTP_PUBKEY' ) ? FTP_PUBKEY : '';
            $credentials['private_key'] = defined( 'FTP_PRIKEY' ) ? FTP_PRIKEY : '';
        }

        if ( in_array( '', $credentials ) ) {
            return false;
        }

        return $credentials;
    }

    /**
     *
     * @param string $id
     * @param bool $page_cache
     *
     * @return string
     */
    private static function _getFileName( $id, $page_cache = false ) {
        // ##<plugin_mod@mirek>#
//        return WPTYPEK_CACHE_DIR . ( $page_cache ? 'page/' : '' ) . md5( NONCE_SALT . $id );

        return WPTYPEK_CACHE_DIR . ( $page_cache ? 'page/' : '' ) . $id;
    }

    /**
     *
     * @throws Exception
     */
    public static function gc() {
        $wp_filesystem = self::getWpFileSystem();

        if ( $wp_filesystem === false ) {
            return false;
        }

        $result = true;

        //Delete any page cache
        $result |= self::deleteCache( 'page' );

        $files = Utility::lsFiles( rtrim( WPTYPEK_CACHE_DIR, '/\\' ), '.', true );
        $now   = time();

        foreach ( $files as $file ) {
            $time = $wp_filesystem->mtime( $file );

            if ( ( $time + self::getLifetime() ) < $now || empty( $time ) ) {
                $result |= $wp_filesystem->delete( $file );
            }
        }

        return $result;
    }

    /**
     * ## <plugin_mod@mirek> ##
     * @throws Exception
     */
    public static function gcCacheOnly() {
        $wp_filesystem = self::getWpFileSystem();

        if ( $wp_filesystem === false ) {
            return false;
        }

        $result = true;

        //Delete any page cache
        $result |= self::deleteCache( 'page' );

        return $result;
    }

    /**
     *
     *
     * @param string $content
     * @param string $id
     * @param bool $page_cache
     *
     * @return string
     * @throws Exception
     */
    public static function saveCache( $content, $id, $page_cache = false ) {
        $wp_filesystem = self::getWpFileSystem();
        //##<plugin_mod@mirek>##

        if ( is_array( $content ) ) {
            $file_contents = serialize( $content );
        } else {
            $file_contents = $content;
        }
        $file = self::_getFileName( $id, $page_cache );

        self::initializeCache();

        if ( $wp_filesystem->put_contents( $file, $file_contents, FS_CHMOD_FILE ) ) {
            if ( $page_cache ) {
                PreloadCache::setCacheTimestamp( get_the_ID() );

                $params      = Plugin::getPluginParams();
                $header_file = str_replace( '.html', '.header', $file );
                if ( $params->get( 'pro_http2_push_enable', '0' ) ) {
                    $aPreloads = Helper::$preloads;
                    if ( ! empty( $aPreloads ) ) {
                        $header_contents = ( 'Link:' . implode( ',', $aPreloads ) );
                        $wp_filesystem->put_contents( $header_file, $header_contents, FS_CHMOD_FILE );
                    }
                }
            }

            return $content;
        } else {
            throw new Exception( __( 'Error writing files to cache' ) );
        }
    }

    protected static function getLifetime( $page_cache = false ) {
        static $lifetime, $page_cache_lifetime;

        if ( $page_cache ) {
            if ( ! $page_cache_lifetime ) {
                $params              = Plugin::getPluginParams();
                $page_cache_lifetime = $params->get( 'page_cache_lifetime', '900' );
            }

            return (int) $page_cache_lifetime;
        }

        if ( ! $lifetime ) {
            $params = Plugin::getPluginParams();

            $lifetime = $params->get( 'cache_lifetime', '900' );
        }

        return (int) $lifetime;
    }
}
