<?php

namespace WpTypek\Core;

defined( '_WPTYPEK_EXEC' ) or die( 'Restricted access' );

use WpTypek\Platform\Plugin;
use WpTypek\Platform\Settings;

class PreloadCache {
    public $enabled = false;
    public $requests;
    public $tableName;
    public $params;

    /**
     * @param Settings $params
     */
    public function __construct() {
        $params          = Plugin::getPluginParams();
        $this->enabled   = (bool) $params->get( 'preload_cache_enable' );
        $this->requests = $params->get( 'preload_cache_requests', 1);
        $this->tableName = static::getTableName();

        add_filter( 'cron_schedules', array( $this, 'addCronInterval' ), 10, 1 );

        add_action( 'admin_init', array( $this, 'createTableIfNotExists' ) );
        add_action( 'save_post', array( $this, 'refreshAllUrls' ) ); //save_post is too often
        add_action( 'delete_post', array( $this, 'refreshAllUrls' ) );
        add_action( 'wptypek_preload_cache_manage_cron_job', array( $this, 'manageCronJob' ) );
        add_action( 'wptypek_preload_cache_cron_job', array( $this, 'createCacheCronJob' ) );

        $this->registerCronJob();
    }

    public function createTableIfNotExists() {
        global $wpdb;

        $tableExists = $wpdb->get_results( "SHOW TABLES LIKE '{$this->tableName}'" );
        if ( ! is_array( $tableExists ) || count( $tableExists ) === 0 ) {
            $query = <<< QUERY
                CREATE TABLE IF NOT EXISTS {$this->tableName} (
                    `id` BIGINT(20) UNSIGNED,
                    `timestamp` TIMESTAMP NULL DEFAULT NULL,
                    PRIMARY KEY (`id`)
                )
QUERY;
            $wpdb->query( $query );
        }
    }

    public function refreshAllUrls() {
        global $wpdb;

        $query         = <<< QUERY
    SELECT ID, post_title, post_type, post_status, guid
    FROM {$wpdb->prefix}posts
    WHERE
        (post_mime_type IS NULL OR post_mime_type = '')
        AND post_status != 'draft'
        AND post_type != 'revision'
    ORDER BY ID ASC
QUERY;
        $rawUrls       = $wpdb->get_results( $query );
        $formattedUrls = [];

        foreach ( $rawUrls as $rawUrl ) {
            try {
                $formattedUrl = strtok( get_permalink( $rawUrl->ID ), '?' );
                if ( empty( $formattedUrls[ $formattedUrl ] ) ) {
                    //OR $id == \WpTypek\Core\Pagecache::getPageCacheId() @todo
                    $formattedUrls[ $formattedUrl ] = $rawUrl->ID;
                }
            } catch ( \Error $e ) {

            }
        }

        $query = "INSERT IGNORE INTO {$this->tableName} (id) VALUES";
        $query .= ( '("' . implode( '"),("', $formattedUrls ) . '")' );

        $wpdb->query( $query );
    }

    public function manageCronJob() {
        if ( $this->enabled ) {
            if ( ! wp_next_scheduled( 'wptypek_preload_cache_cron_job' ) ) {
                $time = time() + 1;
                wp_schedule_single_event( $time, 'wptypek_preload_cache_cron_job' );
            }
        }
    }

    public function createCacheCronJob() {
        if ( $this->enabled ) {
            global $wpdb;

            $tableName          = static::getTableName();
            $expiredTimestamp   = static::getExpiredTimestamp();
            $params             = Plugin::getPluginParams();
            $mobileCacheEnabled = (bool) $params->get( 'mobile_cache_enable', false );
            $limit              = $this->requests;
            if ( $mobileCacheEnabled ) {
                $limit = round($limit / 2);
            }

            $query        = "SELECT * FROM {$tableName} WHERE `timestamp` IS NULL OR `timestamp` < '{$expiredTimestamp}' LIMIT {$limit}";
            $pagesToCache = $wpdb->get_results( $query );
            foreach ( $pagesToCache as $pageToCache ) {
                $permalink = get_permalink( $pageToCache->id );
                wp_remote_get( $permalink, array(
                    'blocking' => false,
                ) );
                if ( $mobileCacheEnabled ) {
                    wp_remote_get( $permalink, array(
                        'blocking'   => false,
                        'user-agent' => 'Mobile',
                    ) );
                }
                PreloadCache::setCacheTimestamp( $pageToCache->id );
            }

            $this->manageCronJob(); //to self-span createCache job
        }
    }

    public function addCronInterval() {
        $schedules['wptypek_every_minute'] = array(
            'interval' => 60,
            'display'  => __( 'Every minute', 'wp-typek' )
        );

        return $schedules;
    }

    public function registerCronJob() {
        if ( $this->enabled && ! wp_next_scheduled( 'wptypek_preload_cache_manage_cron_job' ) ) {
            wp_schedule_event( time(), 'wptypek_every_minute', 'wptypek_preload_cache_manage_cron_job' );
        }
    }

    public static function getTableName() {
        global $wpdb;

        return "{$wpdb->prefix}wptypek_preload_cache";
    }

    public static function setCacheTimestamp( $id, $timestamp = 'NOW()' ) {
        if ( ! empty( $id ) ) {
            global $wpdb;
            $tableName = self::getTableName();
            $query     = <<<QUERY
            INSERT INTO {$tableName} (`id`, `timestamp`) VALUES ({$id}, {$timestamp}) ON DUPLICATE KEY UPDATE `timestamp` = {$timestamp}
QUERY;

            $wpdb->query( $query );
        }
    }

    public static function clearAllCache() {
        global $wpdb;
        $tableName = self::getTableName();
        $query     = <<<QUERY
            UPDATE {$tableName} SET `timestamp` = NULL
QUERY;

        $wpdb->query( $query );
    }

    public static function getPreloadStatus() {
        global $wpdb;
        $tableName        = self::getTableName();
        $expiredTimestamp = static::getExpiredTimestamp();

        $query = <<<QUERY
            SELECT
                COUNT(*) AS 'all',
                COUNT(IF(`timestamp` IS NULL OR `timestamp` < '{$expiredTimestamp}', 1, NULL)) AS 'to_cache',
                COUNT(IF(`timestamp` IS NOT NULL AND `timestamp` >= '{$expiredTimestamp}', 1, NULL)) AS 'cached'
            FROM {$tableName}
QUERY;

        return $wpdb->get_row( $query );

    }

    public static function getExpiredTimestamp() {
        $params                    = Plugin::getPluginParams();
        $pageCacheLifetime         = $params->get( 'page_cache_lifetime', '900' );
        $expiredTimestamp          = new \DateTime();
        $pageCacheLifetimeInterval = new \DateInterval( "PT{$pageCacheLifetime}S" );
        $expiredTimestamp->sub( $pageCacheLifetimeInterval );

        return $expiredTimestamp->format( 'Y-m-d H:i:s' );
    }
}
