<?php

namespace WpTypek\Core;

defined( '_WPTYPEK_EXEC' ) or die( 'Restricted access' );

use WpTypek\Platform\Plugin;
use WpTypek\Platform\Settings;

class ServiceWorker {
    const SERVICE_WORKER_FILENAME = 'pro-service-worker.js';

    public $enabled = false;

    /**
     * @param Settings $params
     */
    public function __construct() {
        $params        = Plugin::getPluginParams();
//        $this->enabled = (bool) $params->get( 'service_worker_enable', 0 );

        add_action( 'plugins_loaded', array( $this, 'setupInterceptUrlHooks' ) );
        add_action( 'wp_enqueue_scripts', array( $this, 'registerServiceWorker' ), 1 );
    }

    public function setupInterceptUrlHooks() {
        add_action( 'init', array( $this, 'interceptRewriteRules' ) );
        add_action( 'parse_request', array( $this, 'checkIfInterceptUrl' ) );
    }

    public function interceptRewriteRules() {
        if ($this->enabled) {
            $serviceWorkerFile = static::SERVICE_WORKER_FILENAME;
            add_rewrite_rule( "^/{$serviceWorkerFile}$",
                "index.php?{$serviceWorkerFile}=1"
            );
        }
    }

    public function checkIfInterceptUrl($query) {
        if ( !$this->enabled || ! property_exists( $query, 'query_vars' ) || ! is_array( $query->query_vars )) {
            return;
        }

        $query_vars_as_string = http_build_query( $query->query_vars );
        if ( strpos( $query_vars_as_string, static::SERVICE_WORKER_FILENAME ) !== false ) {
            $this->outputServiceWorker();
        }
    }

    public function registerServiceWorker() {
        if ( $this->enabled ) {
            wp_enqueue_script( 'wptypek-register-service-worker', WPTYPEK_PLUGIN_URL . 'media/js/pro-register-service-worker.js', array(), null, true );
            $serviceWorkerVersion = WPTYPEK_VERSION;
            $localize             = array(
                'serviceWorkerUrl' => ( site_url( '/' . static::SERVICE_WORKER_FILENAME . '?v=' . $serviceWorkerVersion ) ),
                'cacheName'        => 'wptypek-' . WPTYPEK_VERSION,
                'startPage'        => get_site_url(),
            );

            wp_localize_script( 'wptypek-register-service-worker', 'wptypekServiceWorker', $localize );
        }
    }

    private function outputServiceWorker() {
        header( 'Content-Type: text/javascript' );
        echo $this->getServiceWorkerContent();
        exit();
    }

    private function getServiceWorkerContent() {
        $cacheName = 'wp-typek-' . WPTYPEK_VERSION;
        $url       = get_site_url();

        return <<< SERVICE_WORKER
'use strict';

const cacheName = '$cacheName';
const urlToCache = '$url';
const neverCacheUrls = [/\/wp-admin/,/\/wp-login/,/preview=true/];

// Install
self.addEventListener('install', function(e) {
    console.log('WPTypek service worker installation');
    e.waitUntil(
        caches.open(cacheName).then(function(cache) {
            console.log('WPTypek service worker caching dependencies');
                cache.add(urlToCache).catch(function (reason) {
                    return console.log('WPTypek: ' + String(reason) + ' ' + url);
                });
        })
    );
});

// Activate
self.addEventListener('activate', function(e) {
    console.log('WPTypek service worker activation');
    e.waitUntil(
        caches.keys().then(function(keyList) {

            console.log('cache keyList:', keyList);

            return Promise.all(keyList.map(function(key) {
                if ( key !== cacheName ) {
                    console.log('WPTypek old cache removed', key);
                    return caches.delete(key);
                }
            }));
        })
    );
    return self.clients.claim()
        .then((result) => {
            console.log('WPTypek service ')
        });
});

// Fetch
self.addEventListener('fetch', function(e) {

    // Return if the current request url is in the never cache list
    if ( ! neverCacheUrls.every(checkNeverCacheList, e.request.url) ) {
        console.log( 'WPTypek: Current request is excluded from cache.' );
        return;
    }

    // Return if request url protocol isn't http or https
    if ( ! e.request.url.match(/^(http|https):\/\//i) )
        return;

    // Return if request url is from an external domain.
    if ( new URL(e.request.url).origin !== location.origin )
        return;

    // For POST requests, do not use the cache. Serve offline page if offline.
    if ( e.request.method !== 'GET' ) {
        e.respondWith(
            fetch(e.request).catch( function() {
                return caches.match(offlinePage);
            })
        );
        return;
    }

    // Revving strategy
    if ( e.request.mode === 'navigate' && navigator.onLine ) {
        e.respondWith(
            fetch(e.request).then(function(response) {
                return caches.open(cacheName).then(function(cache) {
                    cache.put(e.request, response.clone());
                    return response;
                });
            })
        );
        return;
    }

    e.respondWith(
        caches.match(e.request).then(function(response) {
            return response || fetch(e.request).then(function(response) {
                return caches.open(cacheName).then(function(cache) {
                    cache.put(e.request, response.clone());
                    return response;
                });
            });
        }).catch(function() {
            return caches.match(offlinePage);
        })
    );
});

// Check if current url is in the neverCacheUrls list
function checkNeverCacheList(url) {
    if ( this.match(url) ) {
        return false;
    }
    return true;
}

SERVICE_WORKER;
    }

    public static function isServiceWorkerRequest() {
        return strpos( $_SERVER['REQUEST_URI'], static::SERVICE_WORKER_FILENAME ) !== false;
    }
}
