<?php

namespace WpTypek\Core;

use WC_Product;
use WP_Post;
use WP_Term;
use WpTypek\Platform\Cache;
use WpTypek\Platform\Plugin;

defined( '_WPTYPEK_EXEC' ) or die( 'Restricted access' );

/**
 * Class WPTypekWordpress
 * ## <plugin_mod@mirek> ##
 */
class WPTypekWooCommerce {
    const WOOCOMMERCE_CACHE_URLS_KEY = 'wptypek_woo_urls';
    const QUANTITY_CHANGE = '0';
    const STOCK_CHANGE = '1';
    const BOTH_CHANGE = '2';
    const NO_CHANGE = '3';
    const PAGE_TYPE_SHOP = 'shop';
    const PAGE_TYPE_PRODUCT = 'product';
    const PAGE_TYPE_CATEGORIES = 'categories';
    const PAGE_TYPE_TAGS = 'tags';
    const MODE_TYPE_QUANTITY = 'q';
    const MODE_TYPE_STOCK = 's';

    private $checkedProducts = [];
    private $params;
    private static $woocommercePluginEnabled = null;

    /**
     * WPTypekWordpress constructor.
     */
    public function __construct() {
        add_action( 'woocommerce_product_set_stock', array( $this, 'onProductStockChange' ) );
        add_action( 'woocommerce_variation_set_stock', array( $this, 'onProductStockChange' ) );

        add_action( 'woocommerce_ajax_added_to_cart', array( $this, 'onProductQuantityChange' ) );
        add_action( 'woocommerce_add_to_cart', array( $this, 'onCartItemQuantityChange' ) );
        add_action( 'woocommerce_remove_cart_item', array( $this, 'onCartItemQuantityChange' ) );
        add_action( 'woocommerce_restore_cart_item', array( $this, 'onCartItemQuantityChange' ) );
        add_action( 'woocommerce_after_cart_item_quantity_update', array( $this, 'onCartItemQuantityChange' ) );

        $this->params = Plugin::getPluginParams();

        /**
         * @TODO - check if additionally handles below hooks
         */
        //do_action( 'woocommerce_checkout_order_processed', $order_id, $posted_data, $order );
        //do_action( 'woocommerce_applied_coupon',  $coupon_code );
        //do_action( 'woocommerce_removed_coupon',  $coupon_code );
    }

    /**
     * @param WC_Product $product
     */
    public function onProductStockChange( $product ) {
        if ( self::isWoocommercePluginEnabled() ) {
            if ( empty( $product ) || is_int( $product ) ) {
                $product = wc_get_product( $product );
            }

            if ( ! empty( $product ) && is_object( $product ) ) {
                $this->checkAndClearCaches( $product, self::MODE_TYPE_STOCK );
            }
        }
    }

    /**
     * @param WC_Product $product
     */
    public function onProductQuantityChange( $product ) {
        if ( self::isWoocommercePluginEnabled() ) {
            if ( empty( $product ) || is_int( $product ) ) {
                $product = wc_get_product( $product );
            }

            if ( ! empty( $product ) && is_object( $product ) ) {
                $this->checkAndClearCaches( $product, self::MODE_TYPE_QUANTITY );
            }
        }
    }

    /**
     * @param string $cart_item_key
     */
    public function onCartItemQuantityChange( $cart_item_key ) {
        if ( self::isWoocommercePluginEnabled() ) {
            global $woocommerce;

            $product = null;
            foreach ( $woocommerce->cart->get_cart() as $from_cart_item_key => $cart_item ) {
                if ( $from_cart_item_key === $cart_item_key ) {
                    $product = wc_get_product( $cart_item['product_id'] );
                }
            }

            if ( ! empty( $product ) && is_object( $product ) ) {
                $this->onProductQuantityChange( $product );
            }
        }
    }

    /**
     * $param WC_Product $product
     *
     * @param WC_Product $product
     * @param string $modeType
     */
    private function checkAndClearCaches( $product, $modeType ) {
        if ( ! isset( $this->checkedProducts[ $product->get_id() ] ) ) {
            $this->checkedProducts[ $product->get_id() ] = true;

            if ( $this->canClearCache( self::PAGE_TYPE_SHOP, $modeType ) ) {
                $this->clearShopCache();
            }

            if ( $this->canClearCache( self::PAGE_TYPE_PRODUCT, $modeType ) ) {
                $this->clearProductCache( $product );
            }

            if ( $this->canClearCache( self::PAGE_TYPE_CATEGORIES, $modeType ) ) {
                $this->clearProductCategoriesCache( $product );
            }

            if ( $this->canClearCache( self::PAGE_TYPE_TAGS, $modeType ) ) {
                $this->clearProductTagsCache( $product );
            }
        }
    }

    /**
     * @param string $pageType
     * @param string $modeType
     *
     * @return bool
     */
    private function canClearCache( $pageType, $modeType ) {
        $selectedChangeOption = $this->params->get( self::buildKey( $pageType ) );

        return (
            $selectedChangeOption === self::BOTH_CHANGE
            || (
                $modeType === self::MODE_TYPE_QUANTITY && $selectedChangeOption === self::QUANTITY_CHANGE
            )
            || (
                $modeType === self::MODE_TYPE_STOCK && $selectedChangeOption === self::STOCK_CHANGE
            )
        );
    }

    /**
     * @param WP_Post|int $cachedObject
     *
     * @return string
     */
    private function getPageCacheIdFor( $cachedObject ) {
        $link = $cachedObject instanceof WP_Term ? get_term_link( $cachedObject ) : get_permalink( $cachedObject );

        return trim( str_replace( '/', WPTYPEK_CACHE_SEPARATOR, str_replace( home_url(), '', $link ) ) . 'cache.html', '-');
    }

    /**
     * @param WP_Post|WP_Term|int $post
     * @param bool $paginated
     *
     * @throws Exception
     */
    private function clearCacheFor( $post, $paginated = false ) {
        $originalPageId = $post->ID;

        if ( ! empty( $originalPageId ) ) {
            PreloadCache::setCacheTimestamp( $originalPageId, 'NULL' );
        }

        $pageCacheId = $this->getPageCacheIdFor( $post );

        $clearCache = function ( $pageCacheId ) {
            $mobilePageCacheId = ( 'mobile' . DIRECTORY_SEPARATOR ) . $pageCacheId;

            $pageCache  = Cache::getCache( $pageCacheId, false, true );
            $cssCacheId = '';
            if ( ! empty( $pageCache ) && preg_match( '/wp\-typek\/css\/(.+)\.css/', $pageCache, $matches ) ) {
                $cssCacheId = $matches[1];
            }
            $mobilePageCache  = Cache::getCache( $pageCacheId, false, true );
            $mobileCssCacheId = '';
            if ( ! empty( $mobilePageCache ) && preg_match( '/wp\-typek\/css\/(.+)\.css/', $mobilePageCache, $matches ) ) {
                $mobileCssCacheId = $matches[1];
            }

            Cache::deleteCssCacheById( $cssCacheId ) && Cache::deleteCssCacheById( $mobileCssCacheId )
            && Cache::deletePagecache( $pageCacheId ) && Cache::deletePagecache( $mobilePageCacheId );
        };

        $clearCache( $pageCacheId );
        if ( $paginated ) {
            $paginatedCacheIdPattern = str_replace( '-cache', '-page-*-cache', $pageCacheId );
            $paginatedCachePaths     = glob( WPTYPEK_CACHE_DIR . 'page/' . $paginatedCacheIdPattern );

            if ( ! empty( $paginatedCachePaths ) && is_array( $paginatedCachePaths ) ) {
                foreach ( $paginatedCachePaths as $paginatedCachePath ) {
                    $paginatedCacheId = str_replace( WPTYPEK_CACHE_DIR . 'page/', '', $paginatedCachePath );
                    $clearCache( $paginatedCacheId );
                }
            }
        }
    }

    private function clearShopCache() {
        $shopPageId = wc_get_page_id( 'shop' );
        if ( ! empty( $shopPageId ) ) {
            $this->clearCacheFor( $shopPageId, true );
        }
    }

    /**
     * @param WC_Product $product
     */
    private function clearProductCache( $product ) {
        $post = get_post( $product->get_id() );
        $this->clearCacheFor( $post );

        if ( $product->has_child() ) {
            foreach ( $product->get_children() as $childProduct ) {
                if ( ! empty( $childProduct ) ) {
                    $childPost = get_post( is_scalar($childProduct) ? $childProduct : $childProduct->get_id() );
                    $this->clearCacheFor( $childPost );
                }
            }
        }
    }

    /**
     * @param WC_Product $product
     */
    private function clearProductCategoriesCache( $product ) {
        $productCategories = get_the_terms( $product->get_id(), 'product_cat' );
        if ( ! empty( $productCategories ) ) {
            foreach ( $productCategories as $productCategory ) {
                if ( ! empty( $productCategory ) ) {
                    $this->clearCacheFor( $productCategory );
                }
            }
        }
    }

    /**
     * @param WC_Product $product
     */
    private function clearProductTagsCache( $product ) {
        $productTags = get_the_terms( $product->get_id(), 'product_tag' );
        if ( ! empty( $productTags ) ) {
            foreach ( $productTags as $productTag ) {
                if ( ! empty( $productTag ) ) {
                    $this->clearCacheFor( $productTag );
                }
            }
        }
    }

    public static function cacheUrls() {
        if ( self::isWoocommercePluginEnabled() ) {
            $urlsToCache = [];

            $urlsToCache[] = self::getPermalinkWithoutProtocol( wc_get_page_id( 'cart' ) );
            $urlsToCache[] = self::getPermalinkWithoutProtocol( wc_get_page_id( 'checkout' ) );
            $wooEndpoints  = WC()->query->get_query_vars();
            $homePage = self::getSiteUrlWithoutProtocol( '/' );

            foreach ( $wooEndpoints as $wooEndpoint ) {
                $wooEndpoint = self::getSiteUrlWithoutProtocol( $wooEndpoint );
                if ( $wooEndpoint != $homePage ) {
                    $urlsToCache[] =  $wooEndpoint;
                }
            }

            update_option( self::WOOCOMMERCE_CACHE_URLS_KEY, $urlsToCache );
        }
    }

    /**
     * @return bool
     */
    public static function isDoingWooCommerceEndpoint() {
        if ( self::isWoocommercePluginEnabled() ) {
            $currentUrl = self::getSiteUrlWithoutProtocol( ( $_SERVER['REQUEST_URI'] ) );
            $wooUrls    = get_option( self::WOOCOMMERCE_CACHE_URLS_KEY );

            return count( array_filter( $wooUrls, function ( $wooUrl ) use ( $currentUrl ) {
                    return strpos( $currentUrl, $wooUrl ) !== false;
                } ) ) > 0;
        }

        return false;
    }

    /**
     * @return bool|null
     */
    public static function isWoocommercePluginEnabled() {
        if ( is_null( self::$woocommercePluginEnabled ) ) {
            $activePlugins                  = get_option( 'active_plugins', [] );
            self::$woocommercePluginEnabled = in_array( 'woocommerce/woocommerce.php', $activePlugins );
        }

        return self::$woocommercePluginEnabled;
    }

    /**
     * @param string $pageType
     *
     * @return string
     */
    public static function buildKey( $pageType ) {
        return "woo_cache_clear_{$pageType}";
    }

    /**
     * @param $postId
     *
     * @return false|string|string[]|\WP_Error
     */
    private static function getPermalinkWithoutProtocol( $postId ) {
        $permalink = get_permalink( $postId );
        if ( ! empty( $permalink ) ) {
            $permalink = str_replace( [ 'https:', 'http:' ], [ '', '' ], $permalink );
        }

        return rtrim( $permalink, '/' ) . '/';
    }

    /**
     * @param string $url
     *
     * @return false|string|string[]|\WP_Error
     */
    private static function getSiteUrlWithoutProtocol( $url ) {
        $siteUrl = site_url( $url );
        if ( ! empty( $siteUrl ) ) {
            $siteUrl = str_replace( [ 'https:', 'http:' ], [ '', '' ], $siteUrl );
        }

        return rtrim( $siteUrl, '/' ) . '/';
    }
}
